from fiscalsim_us.model_api import *


class net_short_term_capital_loss(Variable):
    value_type = float
    entity = TaxUnit
    label = "Net short-term capital loss"
    unit = USD
    documentation = "The excess of short-term capital losses over short-term capital gains."
    definition_period = YEAR
    reference = dict(
        title="26 U.S. Code § 1222(6)",
        href="https://www.law.cornell.edu/uscode/text/26/1222#6",
    )

    formula = excess(
        of="short_term_capital_losses", over="short_term_capital_gains"
    )



from fiscalsim_us.model_api import *


class or_itemized_deductions(Variable):
    value_type = float
    entity = TaxUnit
    label = "OR itemized deductions"
    unit = USD
    definition_period = YEAR
    reference = "https://www.oregonlegislature.gov/bills_laws/ors/ors316.html"  # 316.695 (1)(d)
    defined_for = StateCode.OR

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.irs.deductions
        itm_deds = [
            deduction
            for deduction in p.itemized_deductions
            if deduction not in ["salt_deduction"]
        ]
        or_itemized_deductions_less_salt = add(tax_unit, period, itm_deds)
        property_taxes = add(tax_unit, period, ["real_estate_taxes"])
        salt = p.itemized.salt_and_real_estate
        cap = salt.cap[tax_unit("filing_status", period)]
        capped_property_taxes = min_(property_taxes, cap)
        return or_itemized_deductions_less_salt + capped_property_taxes



from survey_enhance.reweight import LossCategory
from policyengine_core.data import Dataset
import torch
from typing import List, Tuple
import numpy as np
import pandas as pd


def sum_by_household(values: pd.Series, dataset: Dataset) -> np.ndarray:
    return (
        pd.Series(values)
        .groupby(dataset.person.person_household_id.values)
        .sum()
        .values
    )


class Populations(LossCategory):
    weight = 1
    static_dataset = False

    def get_comparisons(
        self, dataset: Dataset
    ) -> List[Tuple[str, float, torch.Tensor]]:
        comparisons = []
        age = dataset.person.age
        comparisons += [
            (
                f"people",
                sum_by_household(np.ones_like(age), dataset),
                330_000_000,
            )
        ]

        return comparisons



from fiscalsim_us.model_api import *


class ny_itemized_deductions_max(Variable):
    value_type = float
    entity = TaxUnit
    label = "NY uncapped itemized deductions"
    unit = USD
    definition_period = YEAR
    reference = "https://www.nysenate.gov/legislation/laws/TAX/615"
    defined_for = StateCode.NY

    def formula(tax_unit, period, parameters):
        gov = parameters(period).gov
        itm_deds = [
            deduction
            for deduction in gov.irs.deductions.itemized_deductions
            if deduction not in ["salt_deduction"]
        ]
        itm_deds_less_salt = add(tax_unit, period, itm_deds)
        property_taxes = add(tax_unit, period, ["real_estate_taxes"])
        salt = gov.irs.deductions.itemized.salt_and_real_estate
        cap = salt.cap[tax_unit("filing_status", period)]
        capped_property_taxes = min_(property_taxes, cap)
        capped_tuition = min_(
            gov.states.ny.tax.income.deductions.itemized.college_tuition_max,
            add(tax_unit, period, ["qualified_tuition_expenses"]),
        )
        return itm_deds_less_salt + capped_property_taxes + capped_tuition



from fiscalsim_us.model_api import *


class me_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Maine deduction"
    unit = USD
    definition_period = YEAR
    reference = "https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_item_stand_%20ded_phaseout_wksht.pdf"
    defined_for = StateCode.ME

    def formula(tax_unit, period, parameters):
        # Line 1. First get their Maine AGI.
        me_agi = tax_unit("me_agi", period)

        # Get filing status.
        filing_status = tax_unit("filing_status", period)

        # Get overall deduction part of parameters tree
        p = parameters(period).gov.states.me.tax.income.deductions.phase_out

        # Calculate the deduction phase-out parameters based on
        # filing status.
        phaseout_start = p.start[filing_status]
        excess = max_(me_agi - phaseout_start, 0)  # Line 3
        phaseout_width = p.width[filing_status]  # Line 4
        phaseout_percent = min_(1, excess / phaseout_width)  # Line 5

        # Get their deduction prior to phaseout. Max of itemized and standard (Line 6)
        max_deduction = max_(
            tax_unit("me_itemized_deductions", period),
            tax_unit("me_standard_deduction", period),
        )

        # Calculate the phaseout amount (Line 7)
        phaseout_amount = max_deduction * phaseout_percent

        return max_deduction - phaseout_amount  # Line 8



from policyengine_core.data import Dataset
import numpy as np
from pathlib import Path
from typing import Type
import pandas as pd
from fiscalsim_us.data.storage import STORAGE_FOLDER
from ..cps import CPS_2023


class CalibratedCPS(Dataset):
    input_dataset: Type[Dataset]
    input_dataset_year: int
    epochs: int = 32
    learning_rate: float = 2e1
    log_dir: str = None
    time_period: str = None
    log_verbose: bool = False

    @staticmethod
    def from_dataset(
        dataset: Type[Dataset],
        new_name: str = None,
        new_label: str = None,
        year: int = None,
        out_year: int = 2023,
        log_folder: str = None,
        verbose: bool = True,
    ):
        class CalibratedCPSFromDataset(CalibratedCPS):
            name = new_name
            label = new_label
            input_dataset = dataset
            input_dataset_year = year or dataset.time_period
            time_period = out_year
            log_dir = log_folder
            file_path = STORAGE_FOLDER / f"{new_name}.h5"
            data_format = dataset.data_format
            log_verbose = verbose

        return CalibratedCPSFromDataset

    def generate(self):
        from .output_cps import OutputDataset
        from survey_enhance.reweight import CalibratedWeights
        from .loss import Loss, calibration_parameters

        input_dataset = OutputDataset.from_dataset(self.input_dataset)()

        original_weights = input_dataset.household.household_weight.values

        calibrated_weights = CalibratedWeights(
            original_weights,
            input_dataset,
            Loss,
            calibration_parameters,
        )
        weights = calibrated_weights.calibrate(
            "2022-01-01",
            epochs=self.epochs,
            learning_rate=self.learning_rate,
            verbose=self.log_verbose,
            log_dir=self.log_dir,
        )

        data = self.input_dataset().load_dataset()

        data["household_weight"] = weights

        self.save_dataset(data)


CalibratedCPS_2023 = CalibratedCPS.from_dataset(
    CPS_2023,
    new_name="calibrated_cps_2023",
    new_label="Calibrated CPS 2023",
    out_year=2023,
)



from fiscalsim_us.model_api import *


class no_salt_income_tax(Variable):
    value_type = float
    entity = TaxUnit
    label = "Federal income tax if SALT were abolished"
    unit = USD
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        simulation = tax_unit.simulation
        no_salt_branch = simulation.get_branch("no_salt")
        no_salt_branch.set_input(
            "salt_deduction", period, np.zeros(tax_unit.count)
        )
        return no_salt_branch.calculate("income_tax", period)



from fiscalsim_us.model_api import *


class ok_eitc(Variable):
    value_type = float
    entity = TaxUnit
    label = "Oklahoma EITC"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/past-year/2021/511-Pkt-2021.pdf"
        "https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-Pkt.pdf"
    )
    defined_for = StateCode.OK

    def formula(tax_unit, period, parameters):
        us_agi = tax_unit("adjusted_gross_income", period)
        ok_agi = tax_unit("ok_agi", period)
        prorate = min_(1, max_(0, where(us_agi != 0, ok_agi / us_agi, 0)))
        us_eitc = tax_unit("earned_income_tax_credit", period)
        p = parameters(period).gov.states.ok.tax.income.credits.earned_income
        return prorate * p.eitc_fraction * us_eitc



from fiscalsim_us.model_api import *


class tax_unit_itemizes(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Itemizes tax deductions"
    unit = USD
    documentation = "Whether this tax unit elects to itemize deductions, rather than claim standard deductions."
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        # First, apply a shortcut. We can't simulate SALT before federal income tax
        # (due to circular dependencies), but we can compare the deduction sizes assuming
        # the SALT is maxed out at its cap.
        ded = parameters(period).gov.irs.deductions
        federal_deductions_if_itemizing = ded.deductions_if_itemizing
        deductions_if_itemizing = [
            deduction
            for deduction in federal_deductions_if_itemizing
            if deduction
            not in [
                "salt_deduction",
                # Exclude QBID to avoid circular reference.
                "qualified_business_income_deduction",
            ]
        ]
        filing_status = tax_unit("filing_status", period)
        salt_cap = ded.itemized.salt_and_real_estate.cap[filing_status]
        itemized_deductions = (
            add(tax_unit, period, deductions_if_itemizing) + salt_cap
        )
        deductions_if_not_itemizing = tax_unit(
            "standard_deduction", period
        )  # Ignore QBID here, it requires SALT.
        if all(itemized_deductions < deductions_if_not_itemizing):
            return False

        tax_if_itemizing = tax_unit("tax_liability_if_itemizing", period)
        tax_if_not_itemizing = tax_unit(
            "tax_liability_if_not_itemizing", period
        )
        return tax_if_itemizing < tax_if_not_itemizing



from fiscalsim_us.model_api import *


class mo_qualified_health_insurance_premiums(Variable):
    value_type = float
    entity = Person
    label = "MO qualified healh insurance premiums"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://dor.mo.gov/forms/5695.pdf",  # Form MO 5695
        "https://www.irs.gov/pub/irs-pdf/f1040sa.pdf",  # Federal Form 1040 Schedule A
        "https://www.irs.gov/pub/irs-pdf/f1040.pdf",  # Federal Form 1040
    )
    defined_for = StateCode.MO

    def formula(person, period, parameters):
        # Variable involves person- and tax unit-level variables.
        tax_unit = person.tax_unit

        # total_health_insurance_premiums is also a primary input to the MO side of calculation, MO Form 5695, Line 8
        # Federal Schedule A, Line 4
        med_expense_deduction = add(
            tax_unit, period, ["medical_expense_deduction"]
        )

        # the ratio of federal medical expense deduction to total medical expenses (out of pocket + premiums)
        # need division because med_dental_out_of_pocket is in federal tax, but no MO tax
        # this ratio is then used to scale the health_insurance_premium amount that can be claimed
        # IRS Schedule A Line 1
        tax_unit_health_expenses = add(
            tax_unit,
            period,
            ["medical_out_of_pocket_expenses", "health_insurance_premiums"],
        )
        # Apply a where statement to avoid division by zero.
        med_expense_deducted_ratio = where(
            tax_unit_health_expenses > 0,
            med_expense_deduction / tax_unit_health_expenses,
            0,
        )

        person_premiums = person("health_insurance_premiums", period)
        tax_unit_premiums = tax_unit.sum(person_premiums)

        # Line 13 of MO Form 5695, represents the portion of medical expenses already deducted via federal tax itemization
        deducted_portion = tax_unit_premiums * med_expense_deducted_ratio
        # subtracts the portion of premiums already deducted from federal tax
        itemized_premiums_amount = tax_unit_premiums - deducted_portion
        itemizes = tax_unit("tax_unit_itemizes", period)
        # Cap at federal taxable income.
        taxable_income = tax_unit("taxable_income", period)
        # Allocate proportionally across people in the tax unit based on share
        # of premiums. Use where statement to avoid dividing by zero.
        person_share_of_tax_unit_premiums = where(
            tax_unit_premiums > 0, person_premiums / tax_unit_premiums, 0
        )

        return person_share_of_tax_unit_premiums * where(
            itemizes,
            min_(itemized_premiums_amount, taxable_income),
            min_(tax_unit_premiums, taxable_income),
        )



import pytest

# Temporarily suspending this test because it is failing on GitHub Actions. The
# failure has to do with the following traceback progression.
# 1. test_microsim.py::test_microsim_runs_cps calls the Microsimulation() class
#    from fiscalsim_us/system.py
# 2. The Microsimulation() class from fiscalsim_us/system.py takes as an input
#    a CoreMicrosimulation  object defined in the Simulation() class of
#    policyengine_core/simulations/simulation.py
# 3. The error is in line 134 of policyengine_core/simulations/simulation.py,
#    which calls the Dataset method from policyengine_core/data/dataset.py
# 4. The error is in line 73 of policyengine_core/data/dataset.py  which uses
#    the download() method in lines 271 to 327. This code tries to download
#    data with a token that we don't have. Although I don't know why it works
#    on my machine. My guess is that we could store the correct datasets on a
#    server of our choice and download those with a different URL using this
#    same method.


@pytest.mark.skipif(True, reason="This test temporarily suspended.")
def test_microsim_runs_cps():
    from fiscalsim_us import Microsimulation

    sim = Microsimulation()
    hnet = sim.calc("household_net_income")
    assert not hnet.isna().any(), "Some households have NaN net income."



from fiscalsim_us.model_api import *


class equiv_household_net_income(Variable):
    value_type = float
    entity = Household
    label = "equivalised net income"
    definition_period = YEAR
    unit = USD
    adds = ["spm_unit_oecd_equiv_net_income"]



from fiscalsim_us.model_api import *


class ok_itemized_deductions(Variable):
    value_type = float
    entity = TaxUnit
    label = "Oklahoma itemized deductions"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/past-year/2021/511-Pkt-2021.pdf"
        "https://oklahoma.gov/content/dam/ok/en/tax/documents/forms/individuals/current/511-Pkt.pdf"
    )
    defined_for = StateCode.OK

    def formula(tax_unit, period, parameters):
        # follows Schedule 511-D in references
        itemizing = tax_unit("tax_unit_itemizes", period)
        # calculate US itemized deductions less state non-property taxes
        us_p = parameters(period).gov.irs.deductions
        items = [
            deduction
            for deduction in us_p.itemized_deductions
            if deduction not in ["salt_deduction"]
        ]
        us_itm_deds_less_salt = add(tax_unit, period, items)
        filing_status = tax_unit("filing_status", period)
        capped_property_taxes = min_(
            add(tax_unit, period, ["real_estate_taxes"]),
            us_p.itemized.salt_and_real_estate.cap[filing_status],
        )
        ok_itm_deds = us_itm_deds_less_salt + capped_property_taxes
        # apply partial limit on OK itemized deductions
        EXEMPT_ITEMS = [
            "medical_expense_deduction",
            "charitable_deduction",
        ]
        exempt_deds = add(tax_unit, period, EXEMPT_ITEMS)
        net_deds = max_(0, ok_itm_deds - exempt_deds)
        ok_p = parameters(period).gov.states.ok.tax.income.deductions
        limited_net_deds = min_(net_deds, ok_p.itemized.limit)
        return itemizing * (exempt_deds + limited_net_deds)



from fiscalsim_us.model_api import *


class me_itemized_deductions(Variable):
    value_type = float
    entity = TaxUnit
    label = "Maine itemized deductions"
    unit = USD
    definition_period = YEAR
    reference = "https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_1040me_sched_2_ff.pdf"
    reference = "https://www.mainelegislature.org/legis/statutes/36/title36sec5125.html"
    defined_for = StateCode.ME



from fiscalsim_us.model_api import *


class nj_agi(Variable):
    value_type = float
    entity = TaxUnit
    label = "New Jersey adjusted gross income"
    unit = USD
    definition_period = YEAR
    reference = "https://law.justia.com/codes/new-jersey/2022/title-54/section-54-8a-36/"
    defined_for = StateCode.NJ

    adds = ["adjusted_gross_income", "nj_agi_additions"]
    subtracts = ["nj_agi_subtractions"]



from fiscalsim_us.model_api import *


class mo_net_state_income_taxes(Variable):
    value_type = float
    entity = TaxUnit
    label = "Missouri net state income taxes"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://dor.mo.gov/forms/MO-A_2021.pdf",
        "https://revisor.mo.gov/main/OneSection.aspx?section=143.141&bid=7212",
    )
    defined_for = StateCode.MO

    def formula(tax_unit, period, parameters):
        filing_status = tax_unit("filing_status", period)
        adjustment_base_amount = parameters(
            period
        ).gov.irs.deductions.itemized.salt_and_real_estate.cap[filing_status]

        # Taxes/income
        state_and_local_sales_or_income_tax = tax_unit(
            "state_and_local_sales_or_income_tax", period
        )

        # Technically, state and local income tax could include the "local_earnings_tax" variable,
        # but because the definition of net_state_income_taxes is simply (state_tax + local_tax) - local_tax,
        # it is redundant to keep it
        # More info about local income tax from the Federal W2 on page 26 here: https://dor.mo.gov/forms/MO-1040%20Instructions_2021.pdf

        net_state_income_taxes = tax_unit("state_income_tax", period)

        income_tax_to_total_ratio = (
            net_state_income_taxes / state_and_local_sales_or_income_tax
        )

        # The threshold is the same as the adjustment_base_amount, i.e. the SALT cap introduced by TCJA
        threshold = adjustment_base_amount

        return where(
            state_and_local_sales_or_income_tax > threshold,
            adjustment_base_amount * (income_tax_to_total_ratio),
            net_state_income_taxes,
        )



from fiscalsim_us.model_api import *


class me_deductions(Variable):
    value_type = float
    entity = TaxUnit
    label = "Maine income deductions"
    unit = USD
    definition_period = YEAR
    reference = ""
    defined_for = StateCode.ME



from .categories.populations import Populations
from survey_enhance.reweight import LossCategory
from policyengine_core.parameters import ParameterNode, uprate_parameters
from pathlib import Path


class Demographics(LossCategory):
    weight = 1
    subcategories = [Populations]


class Loss(LossCategory):
    subcategories = [
        Demographics,
    ]


calibration_parameters = ParameterNode(
    directory_path=Path(__file__).parent.parent / "calibration_parameters",
    name="calibration",
)

calibration_parameters = uprate_parameters(calibration_parameters)



from fiscalsim_us.model_api import *


class ne_taxable_income(Variable):
    value_type = float
    entity = TaxUnit
    label = "NE taxable income"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://revenue.nebraska.gov/files/doc/tax-forms/2021/f_1040n_booklet.pdf"
        "https://revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_5.pdf"
    )
    defined_for = StateCode.NE

    def formula(tax_unit, period, parameters):
        agi = tax_unit("ne_agi", period)
        std_ded = tax_unit("ne_standard_deduction", period)
        itm_ded = tax_unit("ne_itemized_deductions", period)
        deductions = where(itm_ded > std_ded, itm_ded, std_ded)
        return max_(0, agi - deductions)



from fiscalsim_us.model_api import *


class dc_eitc_without_qualifying_child(Variable):
    value_type = float
    entity = TaxUnit
    label = "DC EITC without qualifying children"
    unit = USD
    definition_period = YEAR
    reference = "https://code.dccouncil.gov/us/dc/council/code/sections/47-1806.04"  # (f)
    defined_for = StateCode.DC

    def formula(tax_unit, period, parameters):
        # Start by matching the federal EITC.
        p = parameters(
            period
        ).gov.states.dc.tax.income.credits.eitc.without_children
        federal_eitc = tax_unit("earned_income_tax_credit", period)
        matched_eitc = federal_eitc * p.match
        # Then phase out for income above the phase-out threshold.
        earnings = tax_unit("tax_unit_earned_income", period)
        federal_agi = tax_unit("adjusted_gross_income", period)
        greater_of = max_(earnings, federal_agi)
        excess = max_(greater_of - p.phase_out.start, 0)
        phase_out_amount = excess * p.phase_out.rate
        return max_(federal_eitc - phase_out_amount, 0)



from fiscalsim_us.model_api import *


class eitc_eligible(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Eligible for EITC"
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/32#c_1_A"

    def formula(tax_unit, period, parameters):
        person = tax_unit.members
        has_child = tax_unit("tax_unit_children", period) > 0
        age = person("age", period)
        # Relative parameter reference break branching in some states that
        # modify EITC age limits.
        eitc = parameters.gov.irs.credits.eitc(period)
        min_age = parameters.gov.irs.credits.eitc.eligibility.age.min(period)
        max_age = parameters.gov.irs.credits.eitc.eligibility.age.max(period)
        meets_age_requirements = (age >= min_age) & (age <= max_age)
        no_loss_capital_gains = max_(
            0,
            add(tax_unit, period, ["capital_gains"]),
        )
        eitc_investment_income = (
            add(
                tax_unit,
                period,
                ["net_investment_income", "tax_exempt_interest_income"],
            )
            # Replace limited-loss capital gains with no-loss capital gains.
            - tax_unit("c01000", period)  # Limited-loss capital gains.
            + no_loss_capital_gains
        )
        inv_income_disqualified = (
            eitc_investment_income > eitc.phase_out.max_investment_income
        )
        demographic_eligible = has_child | tax_unit.any(meets_age_requirements)
        # Define eligibility before considering separate filer limitation.
        eligible = demographic_eligible & ~inv_income_disqualified
        # This parameter is true if separate filers are eligible.
        if eitc.eligibility.separate_filer:
            return eligible
        # If separate filers are not eligible, check if the filer is separate.
        filing_status = tax_unit("filing_status", period)
        separate = filing_status == filing_status.possible_values.SEPARATE
        return eligible & ~separate



from fiscalsim_us.model_api import *


class snap_emergency_allotment(Variable):
    value_type = float
    entity = SPMUnit
    definition_period = YEAR
    documentation = "SNAP emergency allotment"
    label = "SNAP emergency allotment"
    reference = "https://www.law.cornell.edu/uscode/text/7/2017#a"
    unit = USD

    def formula(spm_unit, period, parameters):
        snap_ea = 0
        for month in period.get_subperiods("month"):
            snap_ea = snap_ea + spm_unit(
                "snap_emergency_allotment_monthly", month
            )
        return snap_ea



from policyengine_core.data import Dataset
from fiscalsim_us.data.storage import STORAGE_FOLDER
from typing import Type
import pandas as pd
from pathlib import Path


class OutputDataset(Dataset):
    data_format = Dataset.TABLES
    input_dataset: Type[Dataset]
    input_dataset_year: int
    output_year: int
    time_period: int

    @staticmethod
    def from_dataset(
        dataset: Type[Dataset],
        year: int = None,
        out_year: int = 2022,
    ):
        class OutputDatasetFromDataset(OutputDataset):
            name = f"{dataset.name}"
            label = f"{dataset.label}"
            input_dataset = dataset
            time_period = year or dataset.time_period
            output_year = out_year
            file_path = STORAGE_FOLDER / f"output_{dataset.name}.h5"

        return OutputDatasetFromDataset

    def generate(self):
        from fiscalsim_us import Microsimulation

        sim = Microsimulation(
            dataset=self.input_dataset(),
        )

        sim.default_calculation_period = self.output_year

        PERSON_VARIABLES = [
            "age",
            "person_household_id",
        ]

        HOUSEHOLD_VARIABLES = [
            "household_id",
            "household_weight",
        ]

        PROGRAM_VARIABLES = [
            "income_tax",
        ]

        variables = sim.tax_benefit_system.variables

        person = pd.DataFrame()

        for variable in PERSON_VARIABLES:
            person[variable] = sim.calculate(variable, map_to="person").values

        household = pd.DataFrame()

        for variable in HOUSEHOLD_VARIABLES:
            household[variable] = sim.calculate(variable).values

        for variable in PROGRAM_VARIABLES:
            if variables[variable].entity.key != "household":
                person[variable] = sim.calculate(
                    variable, map_to="person"
                ).values
                household[variable] = sim.calculate(
                    variable, map_to="household"
                ).values
                household[f"{variable}_participants"] = sim.map_result(
                    sim.calculate(variable).values > 0,
                    variables[variable].entity.key,
                    "household",
                )
            else:
                household[variable] = sim.calculate(variable).values

        person["person_household_id"] = sim.calculate(
            "household_id", map_to="person"
        ).values

        self.save_dataset(
            dict(
                person=person,
                household=household,
            )
        )



from fiscalsim_us.model_api import *


class md_deductions(Variable):
    value_type = float
    entity = TaxUnit
    label = "MD deductions"
    unit = USD
    definition_period = YEAR
    reference = "https://govt.westlaw.com/mdc/Document/N05479690A64A11DBB5DDAC3692B918BC?viewType=FullText&originationContext=documenttoc&transitionType=CategoryPageItem&contextData=(sc.Default)"
    defined_for = StateCode.MD

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.irs.deductions
        itm_deds = [
            deduction
            for deduction in p.itemized_deductions
            if deduction not in ["salt_deduction"]
        ]
        itm_deds_less_salt = add(tax_unit, period, itm_deds)
        property_taxes = add(tax_unit, period, ["real_estate_taxes"])
        salt = p.itemized.salt_and_real_estate
        cap = salt.cap[tax_unit("filing_status", period)]
        capped_property_taxes = min_(property_taxes, cap)
        md_itm_ded = itm_deds_less_salt + capped_property_taxes
        md_std_ded = tax_unit("md_standard_deduction", period)
        us_itemizer = tax_unit("tax_unit_itemizes", period)
        return where(
            us_itemizer,
            where(md_itm_ded > md_std_ded, md_itm_ded, md_std_ded),
            md_std_ded,
        )



from fiscalsim_us.model_api import *


class me_standard_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Maine standard deduction"
    unit = USD
    definition_period = YEAR
    reference = "https://www.maine.gov/revenue/sites/maine.gov.revenue/files/inline-files/22_item_stand_%20ded_phaseout_wksht.pdf"
    reference = "https://www.mainelegislature.org/legis/statutes/36/title36sec5124-C.html"
    defined_for = StateCode.ME

    def formula(tax_unit, period, parameters):
        # Get filing status.
        filing_status = tax_unit("filing_status", period)

        # Get standard deduction part of parameters tree
        p = parameters(period).gov.states.me.tax.income.deductions.standard

        # Get standard deduction for filing status
        return p.amount[filing_status]



from fiscalsim_us.model_api import *


class mn_standard_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Minnesota standard deduction"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.revenue.state.mn.us/sites/default/files/2023-01/m1_inst_21.pdf"
        "https://www.revenue.state.mn.us/sites/default/files/2023-03/m1_inst_22.pdf"
    )
    defined_for = StateCode.MN

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.mn.tax.income.deductions
        # ... calculate pre-limitation amount
        filing_status = tax_unit("filing_status", period)
        base_amt = p.standard.base[filing_status]
        aged_blind_count = tax_unit("aged_blind_count", period)
        extra_amt = aged_blind_count * p.standard.extra[filing_status]
        std_ded = base_amt + extra_amt
        # ... calculate standard deduction offset
        std_ded_offset = p.deduction_fraction * std_ded
        agi = tax_unit("adjusted_gross_income", period)
        excess_agi = max_(0, agi - p.agi_threshold[filing_status])
        excess_agi_offset = p.excess_agi_fraction * excess_agi
        offset = min_(std_ded_offset, excess_agi_offset)
        return max_(0, std_ded - offset)



from fiscalsim_us.model_api import *


class mo_taxable_income(Variable):
    value_type = float
    entity = Person
    label = "Missouri AGI minus deductions"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://dor.mo.gov/forms/MO-A_2021.pdf",
        "https://www.revisor.mo.gov/main/OneSection.aspx?section=143.111&bid=7201&hl=",
    )
    defined_for = StateCode.MO

    def formula(person, period, parameters):
        # calculate tax unit MO AGI
        tax_unit = person.tax_unit
        mo_agi = person("mo_adjusted_gross_income", period)
        unit_mo_agi = tax_unit.sum(mo_agi)

        # calculate sum of all tax unit MO deductions
        mo_itemized_or_standard = where(
            tax_unit("tax_unit_itemizes", period),  # itemizes on federal form
            tax_unit("mo_itemized_deductions", period),
            tax_unit("standard_deduction", period),  # equal to federal stdded
        )
        mo_federal_income_tax_deduction = tax_unit(
            "mo_federal_income_tax_deduction", period
        )
        mo_pension_and_ss_or_ssd_deduction = tax_unit(
            "mo_pension_and_ss_or_ssd_deduction", period
        )
        unit_mo_deductions = (
            mo_itemized_or_standard
            + mo_federal_income_tax_deduction  # available to all tax units
            + mo_pension_and_ss_or_ssd_deduction  # available to all tax units
        )
        # Note: There would also be a personal and/or dependent exemptions
        # as part of this formula, but they are legally based on eligibility
        # for the federal versions of those exemptions, both of which are
        # suspended through 2025 federally.

        # calculate taxable income for tax unit
        unit_taxinc = max_(0, unit_mo_agi - unit_mo_deductions)

        # allocate unit taxable income by each individual's share of unit AGI
        ind_agi_share = where(unit_mo_agi > 0, mo_agi / unit_mo_agi, 0)
        return ind_agi_share * unit_taxinc



from fiscalsim_us.model_api import *


class ne_itemized_deductions(Variable):
    value_type = float
    entity = TaxUnit
    label = "NE itemized deductions"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://revenue.nebraska.gov/files/doc/tax-forms/2021/f_1040n_booklet.pdf"
        "https://revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_5.pdf"
    )
    defined_for = StateCode.NE

    def formula(tax_unit, period, parameters):
        # 2021 Form 1040N instructions say this:
        #   If you use the standard deduction on the federal return,
        #   you must use the Nebraska standard deduction on the
        #   Nebraska return.  All taxpayers that claimed itemized
        #   deductions on their federal return are allowed the larger
        #   of the Nebraska standard deduction or federal itemized
        #   deductions, minus state and local income taxes claimed on
        #   Federal Schedule A.
        us_itemizing = tax_unit("tax_unit_itemizes", period)
        # calculate US itemized deductions less state non-property taxes
        p = parameters(period).gov.irs.deductions
        items = [
            deduction
            for deduction in p.itemized_deductions
            if deduction not in ["salt_deduction"]
        ]
        us_itm_deds_less_salt = add(tax_unit, period, items)
        filing_status = tax_unit("filing_status", period)
        capped_property_taxes = min_(
            add(tax_unit, period, ["real_estate_taxes"]),
            p.itemized.salt_and_real_estate.cap[filing_status],
        )
        ne_itm_deds = us_itm_deds_less_salt + capped_property_taxes
        return us_itemizing * ne_itm_deds



from fiscalsim_us.model_api import *


class or_federal_tax_liability_subtraction(Variable):
    value_type = float
    entity = TaxUnit
    label = "OR federal tax liability subtraction"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.oregon.gov/dor/forms/FormsPubs/form-or-40-inst_101-040-1_2021.pdf#page=13",
        "https://www.oregonlegislature.gov/bills_laws/ors/ors316.html",  # Subsection 316.800
    )
    defined_for = StateCode.OR

    def formula(tax_unit, period, parameters):
        filing_status = tax_unit("filing_status", period)
        statuses = filing_status.possible_values
        income = tax_unit("adjusted_gross_income", period)
        # Use no-SALT income tax to avoid circular logic, until we add
        # withholding rules.
        federal_tax_liability = tax_unit("no_salt_income_tax", period)
        # Instructions for line 14:
        # "This is your federal income tax liability after refundable credits
        #  (other than the EITC)."
        eitc = tax_unit("eitc", period)
        federal_tax_liability_less_eitc = federal_tax_liability + eitc
        non_negative_federal_tax_liability = max_(
            0, federal_tax_liability_less_eitc
        )
        caps = (
            parameters(period)
            .gov.states["or"]
            .tax.income.subtractions.federal_tax_liability.cap
        )
        cap = select(
            [
                filing_status == statuses.SINGLE,
                filing_status == statuses.JOINT,
                filing_status == statuses.HEAD_OF_HOUSEHOLD,
                filing_status == statuses.SEPARATE,
                filing_status == statuses.WIDOW,
            ],
            [
                caps.single.calc(income),
                caps.joint.calc(income),
                caps.head_of_household.calc(income),
                caps.separate.calc(income),
                caps.widow.calc(income),
            ],
        )
        return min_(non_negative_federal_tax_liability, cap)



from fiscalsim_us.model_api import *


class mn_itemized_deductions(Variable):
    value_type = float
    entity = TaxUnit
    label = "Minnesota itemized deductions"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.revenue.state.mn.us/sites/default/files/2021-12/m1_21_0.pdf"
        "https://www.revenue.state.mn.us/sites/default/files/2023-01/m1_inst_21.pdf"
        "https://www.revenue.state.mn.us/sites/default/files/2022-12/m1_22.pdf"
        "https://www.revenue.state.mn.us/sites/default/files/2023-03/m1_inst_22.pdf"
    )
    defined_for = StateCode.MN

    def formula(tax_unit, period, parameters):
        # 2021 Form M1 instructions say:
        #   You may claim the Minnesota standard deduction or itemize
        #   your deductions on your Minnesota return. You will generally
        #   pay less Minnesota income tax if you take the larger of your
        #   itemized or standard deduction.
        # ... calculate pre-limitation itemized deductions
        p = parameters(period).gov.irs.deductions
        itm_deds = [
            deduction
            for deduction in p.itemized_deductions
            if deduction not in ["salt_deduction"]
        ]
        us_itm_deds_less_salt = add(tax_unit, period, itm_deds)
        filing_status = tax_unit("filing_status", period)
        capped_property_taxes = min_(
            add(tax_unit, period, ["real_estate_taxes"]),
            p.itemized.salt_and_real_estate.cap[filing_status],
        )
        mn_itm_deds = us_itm_deds_less_salt + capped_property_taxes
        # ... calculate itemized deductions offset
        p = parameters(period).gov.states.mn.tax.income.deductions
        exempt_deds = add(
            tax_unit,
            period,
            ["medical_expense_deduction", "casualty_loss_deduction"],
        )
        net_deds = max_(0, mn_itm_deds - exempt_deds)
        net_deds_offset = p.deduction_fraction * net_deds
        agi = tax_unit("adjusted_gross_income", period)
        excess_agi = max_(0, agi - p.agi_threshold[filing_status])
        excess_agi_offset = p.excess_agi_fraction * excess_agi
        offset = min_(net_deds_offset, excess_agi_offset)
        return max_(0, mn_itm_deds - offset)



from fiscalsim_us.model_api import *


class spm_unit_benefits(Variable):
    value_type = float
    entity = SPMUnit
    label = "Benefits"
    definition_period = YEAR
    unit = USD

    def formula(spm_unit, period, parameters):
        BENEFITS = [
            "social_security",
            "ssi",
            "state_supplement",
            "ca_cvrp",  # California Clean Vehicle Rebate Project.
            "snap",
            "wic",
            "free_school_meals",
            "reduced_price_school_meals",
            "lifeline",
            "acp",
            "ebb",
            "tanf",
            "high_efficiency_electric_home_rebate",
            "residential_efficiency_electrification_rebate",
            # Contributed.
            "basic_income",
        ]
        if parameters(period).gov.contrib.ubi_center.flat_tax.deduct_ptc:
            BENEFITS.append("premium_tax_credit")
        if not parameters(period).gov.hud.abolition:
            BENEFITS.append("spm_unit_capped_housing_subsidy")
        return add(spm_unit, period, BENEFITS)



from fiscalsim_us.model_api import *


class cliff_evaluated(Variable):
    value_type = bool
    entity = Person
    label = "cliff evaluated"
    unit = USD
    documentation = "Whether this person's cliff has been simulated. If not, then the cliff gap is assumed to be zero."
    definition_period = YEAR

    def formula(person, period, parameters):
        adult_index_values = person("adult_index", period)
        cliff_adult_count = parameters(period).simulation.cliff_adults
        return adult_index_values <= cliff_adult_count


class cliff_gap(Variable):
    value_type = float
    entity = Person
    label = "cliff gap"
    unit = USD
    documentation = "Amount of income lost if this person's employment income increased by $2,000."
    definition_period = YEAR

    def formula(person, period, parameters):
        simulation = person.simulation
        alt_simulation = simulation.get_branch(f"all_adults_2k_pay_rise")
        delta = parameters(period).simulation.cliff_threshold
        for variable in simulation.tax_benefit_system.variables:
            if variable not in simulation.input_variables:
                alt_simulation.delete_arrays(variable)
        alt_simulation.set_input(
            "employment_income",
            period,
            person("employment_income", period)
            + person("is_adult", period) * delta,
        )
        alt_person = alt_simulation.person
        household_net_income = person.spm_unit("spm_unit_net_income", period)
        household_net_income_higher_earnings = alt_person.spm_unit(
            "spm_unit_net_income", period
        )
        increase = household_net_income_higher_earnings - household_net_income
        return where(person("is_adult", period) & (increase < 0), -increase, 0)


class is_on_cliff(Variable):
    value_type = bool
    entity = Person
    label = "is on a tax-benefit cliff"
    documentation = "Whether this person would be worse off if their employment income were $2,000 higher."
    definition_period = YEAR

    def formula(person, period, parameters):
        return person("cliff_gap", period) > 0



from .cps import (
    CPS_2020,
    CPS_2021,
    CPS_2022,
    CPS_2023,
    RawCPS_2020,
    RawCPS_2021,
    CalibratedCPS_2023,
)

DATASETS = [CPS_2020, CPS_2021, CPS_2022, CPS_2023, CalibratedCPS_2023]
